from __future__ import annotations

from collections.abc import Hashable
from weakref import WeakValueDictionary


class HashableWrapper(Hashable):
    """Wrapper for objects that are not hashable.

    It will stay alive as long as the object is alive.
    """

    __slots__ = ("key", "obj", "__weakref__")

    def __init__(self, obj: object):
        self.key = id(obj)
        self._track(obj)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, HashableWrapper):
            return self.key == other.key
        return NotImplemented

    def __hash__(self) -> int:
        return hash(self.key)

    def _track(self, obj: object) -> None:
        if self in _wrapper_to_obj:
            assert _wrapper_to_obj[self] is obj, (
                "There is already a wrapper for the provided object, but it's "
                "pointing to a different object. This is probably a bug, "
                "please report it."
            )
            return
        _wrapper_to_obj[self] = obj


_wrapper_to_obj: WeakValueDictionary[HashableWrapper, object] = WeakValueDictionary()



from __future__ import annotations

import math
import sys
from typing import Callable, Dict, Hashable, MutableMapping, Tuple, TypeVar, overload
from weakref import WeakKeyDictionary

import cachetools

from ._hashable import HashableWrapper

if sys.version_info >= (3, 10):
    from typing import ParamSpec, TypeAlias
else:
    from typing_extensions import ParamSpec, TypeAlias


P = ParamSpec("P")
T = TypeVar("T")


MethodCache: TypeAlias = MutableMapping

CacheFactory = Callable[[], MethodCache]

ObjectCache: TypeAlias = Dict[Callable[P, T], MethodCache[Tuple[Hashable], T]]

_cache_by_object: WeakKeyDictionary[object, ObjectCache] = WeakKeyDictionary()


def default_cache_factory() -> MethodCache:
    return cachetools.Cache(maxsize=math.inf)


def get_cache(
    obj: object, method: Callable[P, T], cache_factory: CacheFactory | None = None
) -> MutableMapping[tuple[Hashable], T]:
    w = HashableWrapper(obj)
    try:
        instance_cache = _cache_by_object[w]
    except KeyError:
        # TODO: Should we use a WeakRefDictionary here too?
        instance_cache = _cache_by_object[w] = {}

    try:
        method_cache = instance_cache[method]
    except KeyError:
        if cache_factory is None:
            raise TypeError(
                "The cache for the provided object and method was never initialized"
                "and it can't be created, since `cache_factory` param is None"
            )
        method_cache = instance_cache[method] = cache_factory()
    return method_cache


@overload
def cached_method(
    *, cache_factory: CacheFactory = ...
) -> Callable[[Callable[P, T]], Callable[P, T]]:
    ...


@overload
def cached_method(
    method: Callable[P, T], *, cache_factory: CacheFactory = ...
) -> Callable[P, T]:
    ...


def cached_method(
    method: Callable[P, T] | None = None,
    *,
    cache_factory: CacheFactory = default_cache_factory,
) -> Callable[P, T] | Callable[[Callable[P, T]], Callable[P, T]]:
    if method is None:

        def wrapper(method: Callable[P, T]) -> Callable[P, T]:
            return cached_method(method, cache_factory=cache_factory)

        return wrapper

    def cache_getter(obj: object) -> MethodCache:
        slots = getattr(type(obj), "__slots__", None)
        if slots is not None and "__weakref__" not in slots:
            raise TypeError(
                "In order for `cached_method` to support classes with __slots__, "
                'you need to add the "__weakref__" attribute to the __slots__'
            )

        assert method is not None  # for mypy
        return get_cache(obj, method, cache_factory)

    # TODO: Add support to override the `lock` and `key` param
    return cachetools.cachedmethod(cache=cache_getter)(method)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['methodic_cache']

package_data = \
{'': ['*']}

install_requires = \
['cachetools>=5.0.0,<6.0.0']

extras_require = \
{':python_version < "3.10"': ['typing-extensions>=4.5.0,<5.0.0']}

setup_kwargs = {
    'name': 'methodic-cache',
    'version': '0.2.1',
    'description': 'functools.cache() for methods, done correctly',
    'long_description': '# methodic_cache\n`functools.cache()` for methods, done correctly.\n\n# Features\n* Simple to use\n* Extendable with [custom cache backends](#custom-cache-backends) (e.g. LRUCache, LFUCache, etc.)\n* Works with non-hashable objects\n* Works with [frozen/slotted classes](#using-classes-with-__slots__)\n* No memory leaks\n* 100% test coverage\n\n# Installation\n```bash\npip install methodic_cache\n```\n\n# Usage\n```python\nfrom methodic_cache import cached_method\n\n\nclass MyClass:\n    @cached_method\n    def my_method(self, arg1, arg2):\n        return arg1 + arg2\n\n\nmy_obj = MyClass()\nmy_obj.my_method(1, 2)  # returns 3\nmy_obj.my_method(1, 2)  # returns 3 from the cache\n```\n\n\n## Using classes with `__slots__`\nClasses that define `__slots__` need to have a `__weakref__` slot to be able to be weakly referenced:\n\n```python\nfrom methodic_cache import cached_method\n\n\nclass MyClass:\n    __slots__ = ("my_attr", "__weakref__")  # <-- __weakref__ is required\n\n    def __init__(self, my_attr):\n        self.my_attr = my_attr\n\n    @cached_method\n    def my_method(self, arg1, arg2):\n        print(f"Computing {self.my_attr} + {arg1} + {arg2}...")\n        return self.my_attr + arg1 + arg2\n\nmy_obj = MyClass(1)\nmy_obj.my_method(2, 3)\n# prints "Computing 1 + 2 + 3..."\n# returns 6\nmy_obj.my_method(2, 3)\n# returns 6\n```\n\n\n## Custom cache backends\nYou can use any cache backend that implements the `MutableMapping` interface (e.g. `dict`, `lru_cache`, `functools.lru_cache`, etc.).\nThe default cache backend is `cachetools.Cache(maxsize=math.inf)`, which will keep the cache bounded to the lifetime of the `self` object.\n\nYou can use a different cache backend by passing it as the `cache_factory` argument to `cached_method`:\n\n```python\nfrom methodic_cache import cached_method\nfrom cachetools import LRUCache\n\n\nclass MyClass:\n    @cached_method(cache_factory=lambda: LRUCache(maxsize=1))\n    def my_method(self, arg1, arg2):\n        print(f"Computing {arg1} + {arg2}...")\n        return arg1 + arg2\n\n\nmy_obj = MyClass()\nmy_obj.my_method(1, 1)\n# prints Computing 1 + 1...\n# returns 2\nmy_obj.my_method(1, 1)\n# returns 2\nmy_obj.my_method(2, 2)\n# prints Computing 2 + 2...\n# returns 4\nmy_obj.my_method(1, 1)  # <-- this will be recomputed because the cache is full\n# prints Computing 1 + 1...\n# returns 2\n```\n',
    'author': 'Alessio Bogon',
    'author_email': '778703+youtux@users.noreply.github.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/youtux/methodic_cache',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
